"""
Module with global fixtures
"""
import os

import boto3
from botocore.exceptions import ClientError, NoCredentialsError
import pytest

from sentinelhub import SHConfig
from eolearn.core import EOPatch

EXAMPLE_DATA_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..', '..', 'example_data')
TEST_EOPATCH_PATH = os.path.join(EXAMPLE_DATA_PATH, 'TestEOPatch')


@pytest.fixture(name='test_eopatch')
def test_eopatch_fixture():
    return EOPatch.load(TEST_EOPATCH_PATH, lazy_loading=True)


@pytest.fixture(name='example_data_path')
def example_data_path_fixture():
    return EXAMPLE_DATA_PATH


@pytest.fixture(name='config')
def config_fixture():
    config = SHConfig()
    for param in config.get_params():
        env_variable = param.upper()
        if os.environ.get(env_variable):
            setattr(config, param, os.environ.get(env_variable))
    return config


@pytest.fixture(name='gpkg_file')
def local_gpkg_example_file_fixture():
    """ A pytest fixture to retrieve a gpkg example file
    """
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../../../example_data/import-gpkg-test.gpkg')
    return path


@pytest.fixture(name='s3_gpkg_file')
def s3_gpkg_example_file_fixture(config):
    """ A pytest fixture to retrieve a gpkg example file
    """
    aws_config = {
        "region_name": 'eu-central-1',
    }
    if config.aws_access_key_id and config.aws_secret_access_key:
        aws_config['aws_access_key_id'] = config.aws_access_key_id
        aws_config['aws_secret_access_key'] = config.aws_secret_access_key

    client = boto3.client('s3', **aws_config)

    try:
        client.head_bucket(Bucket='eolearn-io')
    except (ClientError, NoCredentialsError):
        return pytest.skip(msg='No access to the bucket.')

    return 's3://eolearn-io/import-gpkg-test.gpkg'


@pytest.fixture(name='geodb_client')
def geodb_client_fixture():
    """ A geoDB client object
    """
    geodb = pytest.importorskip('xcube_geodb.core.geodb')

    client_id = os.getenv('GEODB_AUTH_CLIENT_ID')
    client_secret = os.getenv('GEODB_AUTH_CLIENT_SECRET')

    if not (client_id or client_secret):
        raise ValueError("Could not initiate geoDB client, GEODB_AUTH_CLIENT_ID and GEODB_AUTH_CLIENT_SECRET missing!")

    return geodb.GeoDBClient(
        client_id=client_id,
        client_secret=client_secret
    )



"""
Credits:
Copyright (c) 2021-> Matej Aleksandrov, Matej Batič (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import pytest

from eolearn.core import FeatureType
from eolearn.io import GeoDBVectorImportTask, GeopediaVectorImportTask, VectorImportTask
from sentinelhub import BBox, CRS


@pytest.mark.parametrize(
    argnames='reproject, clip, n_features, bbox, crs',
    ids=['simple', 'bbox', 'bbox_full', 'bbox_smaller'],
    argvalues=[
        (False, False, 193, None, None),
        (False, False, 193, BBox([857000, 6521500, 861000, 6525500], CRS('epsg:2154')), None),
        (True, True, 193, BBox([657089, 5071037, 661093, 5075039], CRS.UTM_31N), CRS.UTM_31N),
        (True, True, 125, BBox([657690, 5071637, 660493, 5074440], CRS.UTM_31N), CRS.UTM_31N)
    ])
class TestVectorImportTask:
    """ Class for testing vector imports from local file, s3 bucket object and layer from Geopedia
    """
    def test_import_local_file(self, gpkg_file, reproject, clip, n_features, bbox, crs):
        self._test_import(bbox, clip, crs, gpkg_file, n_features, reproject)

    def test_import_s3_file(self, s3_gpkg_file, reproject, clip, n_features, bbox, crs):
        self._test_import(bbox, clip, crs, s3_gpkg_file, n_features, reproject)

    def test_import_from_geopedia(self, reproject, clip, n_features, bbox, crs):
        feature = FeatureType.VECTOR_TIMELESS, 'lpis_iacs'
        import_task = GeopediaVectorImportTask(
            feature=feature,
            geopedia_table=3447,
            reproject=reproject,
            clip=clip
        )
        eop = import_task.execute(bbox=bbox)
        assert len(eop[feature]) == n_features, 'Wrong number of features!'
        to_crs = crs or import_task.dataset_crs
        assert eop[feature].crs.to_epsg() == to_crs.epsg

    @staticmethod
    def _test_import(bbox, clip, crs, gpkg_example, n_features, reproject):
        feature = FeatureType.VECTOR_TIMELESS, 'lpis_iacs'
        import_task = VectorImportTask(
            feature=feature,
            path=gpkg_example,
            reproject=reproject,
            clip=clip
        )
        eop = import_task.execute(bbox=bbox)
        assert len(eop[feature]) == n_features, 'Wrong number of features!'
        to_crs = crs or import_task.dataset_crs
        assert eop[feature].crs == to_crs.pyproj_crs()


def test_clipping_wrong_crs(gpkg_file):
    """Test for trying to clip using different CRS than the data is in"""
    with pytest.raises(ValueError):
        feature = FeatureType.VECTOR_TIMELESS, 'lpis_iacs'
        import_task = VectorImportTask(
            feature=feature,
            path=gpkg_file,
            reproject=False,
            clip=True
        )
        import_task.execute(bbox=BBox([657690, 5071637, 660493, 5074440], CRS.UTM_31N))


@pytest.mark.parametrize(
    argnames='reproject, clip, n_features, bbox',
    ids=['bbox', 'bbox_smaller'],
    argvalues=[
        (False, False, 227, BBox([857000, 6521500, 861000, 6525500], crs='epsg:2154')),
        (True, True, 162, BBox([857400, 6521900, 860600, 6525100], crs='epsg:2154'))
    ])
def test_import_from_geodb(geodb_client, reproject, clip, n_features, bbox):
    """Test for importing from GeoDB.
    It will only run if geodb credentials are available as the environment variables
    """
    assert geodb_client.whoami, 'Client is not set-up correctly'

    feature = FeatureType.VECTOR_TIMELESS, 'lpis_iacs'
    import_task = GeoDBVectorImportTask(
        feature=feature,
        geodb_client=geodb_client,
        geodb_collection='france_metropolitan_gsaa_2015',
        geodb_db='lpis_iacs',
        reproject=reproject,
        clip=clip
    )
    eop = import_task.execute(bbox=bbox)
    assert len(eop[feature]) == n_features, 'Wrong number of features!'
    assert eop[feature].crs.to_epsg() == eop.bbox.crs.epsg



"""
Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)
Copyright (c) 2018-2019 William Ouellette
Copyright (c) 2019 Drew Bollinger (DevelopmentSeed)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import copy
import datetime
import logging
import os
import tempfile
import dataclasses
from typing import Optional, Union

import boto3
import numpy as np
from numpy.testing import assert_array_equal
from fs.errors import ResourceNotFound
from moto import mock_s3
import pytest

from eolearn.core import EOPatch, FeatureType
from eolearn.io import ExportToTiffTask, ImportFromTiffTask
from sentinelhub import read_data
from sentinelhub.time_utils import serialize_time

from conftest import TEST_EOPATCH_PATH

logging.basicConfig(level=logging.DEBUG)


BUCKET_NAME = 'mocked-test-bucket'
PATH_ON_BUCKET = f's3://{BUCKET_NAME}/some-folder'


@pytest.fixture(autouse=True)
def create_s3_bucket_fixture():
    with mock_s3():
        s3resource = boto3.resource('s3', region_name='eu-central-1')
        s3resource.create_bucket(Bucket=BUCKET_NAME, CreateBucketConfiguration={'LocationConstraint': 'eu-central-1'})
        yield


@dataclasses.dataclass
class TiffTestCase:
    name: str
    feature_type: FeatureType
    data: np.ndarray
    bands: Optional[Union[tuple, list]] = None
    times: Optional[Union[tuple, list]] = None
    expected_times: Optional[Union[tuple, list]] = None

    def __post_init__(self):
        if self.expected_times is None:
            self.expected_times = self.times

    def get_expected(self):
        """ Returns expected data at the end of export-import process
        """
        expected = self.data.copy()

        if isinstance(self.expected_times, tuple):
            expected = expected[self.expected_times[0]: self.expected_times[1] + 1, ...]
        elif isinstance(self.expected_times, list):
            expected = expected[self.expected_times, ...]

        if isinstance(self.bands, tuple):
            expected = expected[..., self.bands[0]: self.bands[1] + 1]
        elif isinstance(self.bands, list):
            expected = expected[..., self.bands]

        if expected.dtype == np.int64:
            expected = expected.astype(np.int32)

        return expected

    def get_expected_timestamp_size(self):
        if self.feature_type.is_timeless():
            return None
        return self.get_expected().shape[0]


DATES = EOPatch.load(TEST_EOPATCH_PATH, lazy_loading=True).timestamp
SCALAR_ARRAY = np.arange(10 * 6, dtype=np.float32).reshape(10, 6)
MASK_ARRAY = np.arange(5*3*2*1, dtype=np.uint16).reshape(5, 3, 2, 1)
DATA_TIMELESS_ARRAY = np.arange(3*2*5, dtype=np.float64).reshape(3, 2, 5)
DATA_ARRAY = np.arange(10 * 3 * 2 * 6, dtype=np.float32).reshape(10, 3, 2, 6)


TIFF_TEST_CASES = [
    TiffTestCase('scalar_timeless', FeatureType.SCALAR_TIMELESS, np.arange(3)),
    TiffTestCase('scalar_timeless_list', FeatureType.SCALAR_TIMELESS, np.arange(5), bands=[3, 0, 2]),
    TiffTestCase('scalar_timeless_tuple', FeatureType.SCALAR_TIMELESS, np.arange(6), bands=(1, 4)),
    TiffTestCase('scalar_band_single_time_single', FeatureType.SCALAR, SCALAR_ARRAY, bands=[3], times=[7]),
    TiffTestCase('scalar_band_list_time_list', FeatureType.SCALAR, SCALAR_ARRAY,
                 bands=[2, 4, 1, 0], times=[1, 7, 0, 2, 3]),
    TiffTestCase('scalar_band_tuple_time_tuple', FeatureType.SCALAR, SCALAR_ARRAY, bands=(1, 4), times=(2, 8)),
    TiffTestCase('mask_timeless', FeatureType.MASK_TIMELESS, np.arange(3*3*1).reshape(3, 3, 1)),
    TiffTestCase('mask_single', FeatureType.MASK, MASK_ARRAY, times=[4]),
    TiffTestCase('mask_list', FeatureType.MASK, MASK_ARRAY, times=[4, 2]),
    TiffTestCase('mask_tuple_int', FeatureType.MASK, MASK_ARRAY, times=(2, 4)),
    TiffTestCase('mask_tuple_datetime', FeatureType.MASK, MASK_ARRAY, times=(DATES[2], DATES[4]),
                 expected_times=(2, 4)),
    TiffTestCase('mask_tuple_string', FeatureType.MASK, MASK_ARRAY, times=(serialize_time(DATES[2]),
                 serialize_time(DATES[4])), expected_times=(2, 4)),
    TiffTestCase('data_timeless_band_list', FeatureType.DATA_TIMELESS, DATA_TIMELESS_ARRAY, bands=[2, 4, 1, 0]),
    TiffTestCase('data_timeless_band_tuple', FeatureType.DATA_TIMELESS, DATA_TIMELESS_ARRAY, bands=(1, 4)),
    TiffTestCase('data_band_list_time_list', FeatureType.DATA, DATA_ARRAY, bands=[2, 4, 1, 0], times=[1, 7, 0, 2, 3]),
    TiffTestCase('data_band_tuple_time_tuple', FeatureType.DATA, DATA_ARRAY, bands=(1, 4), times=(2, 8)),
    TiffTestCase('data_normal', FeatureType.DATA, DATA_ARRAY),
]


@pytest.mark.parametrize('test_case', TIFF_TEST_CASES)
def test_export_import(test_case, test_eopatch):

    test_eopatch[test_case.feature_type][test_case.name] = test_case.data

    with tempfile.TemporaryDirectory() as tmp_dir_name:
        tmp_file_name = 'temp_file.tiff'
        tmp_file_name_reproject = 'temp_file_4326.tiff'
        feature = test_case.feature_type, test_case.name

        export_task = ExportToTiffTask(
            feature, folder=tmp_dir_name, band_indices=test_case.bands, date_indices=test_case.times
        )
        export_task(test_eopatch, filename=tmp_file_name)

        export_task = ExportToTiffTask(
            feature, folder=tmp_dir_name, band_indices=test_case.bands, date_indices=test_case.times,
            crs='EPSG:4326', compress='lzw'
        )
        export_task(test_eopatch, filename=tmp_file_name_reproject)

        import_task = ImportFromTiffTask(
            feature, folder=tmp_dir_name, timestamp_size=test_case.get_expected_timestamp_size()
        )

        expected_raster = test_case.get_expected()

        new_eop = import_task(filename=tmp_file_name)
        old_eop = import_task(test_eopatch, filename=tmp_file_name)

        assert_array_equal(expected_raster, new_eop[test_case.feature_type][test_case.name],
                           err_msg='Tiff imported into new EOPatch is not the same as expected')
        assert_array_equal(expected_raster, old_eop[test_case.feature_type][test_case.name],
                           err_msg='Tiff imported into old EOPatch is not the same as expected')
        assert expected_raster.dtype == new_eop[test_case.feature_type][test_case.name].dtype, \
            'Tiff imported into new EOPatch has different dtype as expected'


@pytest.mark.parametrize('bands, times', [
    ([2, 'string', 1, 0], [1, 7, 0, 2, 3]),
    ([2, 3, 1, 0], [1, 7, 'string', 2, 3])
])
def test_export2tiff_wrong_format(bands, times, test_eopatch):
    test_eopatch.data['data'] = np.arange(10*3*2*6, dtype=float).reshape(10, 3, 2, 6)

    with tempfile.TemporaryDirectory() as tmp_dir_name, pytest.raises(ValueError):
        tmp_file_name = 'temp_file.tiff'
        task = ExportToTiffTask(
            (FeatureType.DATA, 'data'), folder=tmp_dir_name, band_indices=bands, date_indices=times, image_dtype=float
        )
        task.execute(test_eopatch, filename=tmp_file_name)


def test_export2tiff_wrong_feature(mocker, test_eopatch):
    mocker.patch('logging.Logger.warning')

    with tempfile.TemporaryDirectory() as tmp_dir_name:
        tmp_file_name = 'temp_file.tiff'
        feature = FeatureType.MASK_TIMELESS, 'feature-not-present'

        export_task = ExportToTiffTask(feature, folder=tmp_dir_name, fail_on_missing=False)
        export_task(test_eopatch, filename=tmp_file_name)

        assert logging.Logger.warning.call_count == 1

        (val_err,), _ = logging.Logger.warning.call_args
        assert str(val_err) == 'Feature feature-not-present of type FeatureType.MASK_TIMELESS was not found in EOPatch'

        with pytest.raises(ValueError):
            failing_export_task = ExportToTiffTask(feature, folder=tmp_dir_name, fail_on_missing=True)
            failing_export_task(test_eopatch, filename=tmp_file_name)


def test_export2tiff_separate_timestamps(test_eopatch):
    test_case = TIFF_TEST_CASES[-1]
    eopatch = copy.deepcopy(test_eopatch)
    eopatch[test_case.feature_type][test_case.name] = test_case.data
    eopatch.timestamp = test_eopatch.timestamp[:test_case.data.shape[0]]

    with tempfile.TemporaryDirectory() as tmp_dir_name:
        tmp_file_name = 'temp_file_*'
        tmp_file_name_reproject = 'temp_file_4326_%Y%m%d.tif'
        feature = test_case.feature_type, test_case.name

        export_task = ExportToTiffTask(feature, band_indices=test_case.bands, date_indices=test_case.times)
        full_path = os.path.join(tmp_dir_name, tmp_file_name)
        export_task(eopatch, filename=full_path)

        for timestamp in eopatch.timestamp:
            expected_path = os.path.join(tmp_dir_name, timestamp.strftime('temp_file_%Y%m%dT%H%M%S.tif'))
            assert os.path.exists(expected_path), f'Path {expected_path} does not exist'

        full_path = os.path.join(tmp_dir_name, tmp_file_name_reproject)
        export_task = ExportToTiffTask(
            feature, folder=full_path, band_indices=test_case.bands, date_indices=test_case.times, crs='EPSG:4326',
            compress='lzw'
        )
        export_task(eopatch)

        for timestamp in eopatch.timestamp:
            expected_path = os.path.join(tmp_dir_name, timestamp.strftime(tmp_file_name_reproject))
            assert os.path.exists(expected_path), f'Path {expected_path} does not exist'


def test_import_tiff_subset(test_eopatch, example_data_path):
    path = os.path.join(example_data_path, 'import-tiff-test1.tiff')
    mask_feature = FeatureType.MASK_TIMELESS, 'TEST_TIF'

    task = ImportFromTiffTask(mask_feature, path)
    task(test_eopatch)

    tiff_img = read_data(path)

    assert_array_equal(tiff_img[20: 53, 21: 54], test_eopatch[mask_feature][..., 0],
                       err_msg='Imported tiff data should be the same as original')


def test_import_tiff_intersecting(test_eopatch, example_data_path):
    path = os.path.join(example_data_path, 'import-tiff-test2.tiff')
    mask_feature = FeatureType.MASK_TIMELESS, 'TEST_TIF'
    no_data_value = 1.0

    task = ImportFromTiffTask(mask_feature, path, image_dtype=np.float64, no_data_value=no_data_value)
    task(test_eopatch)

    tiff_img = read_data(path)

    assert_array_equal(tiff_img[-6:, :3, :], test_eopatch[mask_feature][:6, -3:, :],
                       err_msg='Imported tiff data should be the same as original')
    feature_dtype = test_eopatch[mask_feature].dtype
    assert feature_dtype == np.float64, f'Feature should have dtype numpy.float64 but {feature_dtype} found'

    test_eopatch[mask_feature][:6, -3:, :] = no_data_value
    unique_values = list(np.unique(test_eopatch[mask_feature][:6, -3:, :]))
    assert unique_values == [no_data_value], f'No data values should all be equal to {no_data_value}'


def test_timeless_feature(test_eopatch):
    feature = FeatureType.DATA_TIMELESS, 'DEM'
    filename = 'relative-path/my-filename.tiff'

    export_task = ExportToTiffTask(feature, folder=PATH_ON_BUCKET)
    import_task = ImportFromTiffTask(feature, folder=PATH_ON_BUCKET)

    export_task.execute(test_eopatch, filename=filename)
    new_eopatch = import_task.execute(test_eopatch, filename=filename)

    assert_array_equal(new_eopatch[feature], test_eopatch[feature])


def test_time_dependent_feature(test_eopatch):
    feature = FeatureType.DATA, 'NDVI'
    filename_export = 'relative-path/*.tiff'
    filename_import = [f'relative-path/{timestamp.strftime("%Y%m%dT%H%M%S")}.tiff'
                       for timestamp in test_eopatch.timestamp]

    export_task = ExportToTiffTask(feature, folder=PATH_ON_BUCKET)
    import_task = ImportFromTiffTask(feature, folder=PATH_ON_BUCKET, timestamp_size=68)

    export_task(test_eopatch, filename=filename_export)
    new_eopatch = import_task(filename=filename_import)

    assert_array_equal(new_eopatch[feature], test_eopatch[feature])

    test_eopatch.timestamp[-1] = datetime.datetime(2020, 10, 10)
    filename_import = [f'relative-path/{timestamp.strftime("%Y%m%dT%H%M%S")}.tiff'
                       for timestamp in test_eopatch.timestamp]

    with pytest.raises(ResourceNotFound):
        import_task(filename=filename_import)


def test_time_dependent_feature_with_timestamps(test_eopatch):
    feature = FeatureType.DATA, 'NDVI'
    filename = 'relative-path/%Y%m%dT%H%M%S.tiff'

    export_task = ExportToTiffTask(feature, folder=PATH_ON_BUCKET)
    import_task = ImportFromTiffTask(feature, folder=PATH_ON_BUCKET)

    export_task.execute(test_eopatch, filename=filename)
    new_eopatch = import_task(test_eopatch, filename=filename)

    assert_array_equal(new_eopatch[feature], test_eopatch[feature])



"""
A collection of input and output EOTasks
"""

from .geopedia import AddGeopediaFeature, AddGeopediaFeatureTask
from .local_io import ExportToTiff, ImportFromTiff, ExportToTiffTask, ImportFromTiffTask
from .geometry_io import VectorImportTask, GeopediaVectorImportTask, GeoDBVectorImportTask
from .sentinelhub_process import SentinelHubDemTask, SentinelHubEvalscriptTask, SentinelHubInputTask, \
    SentinelHubSen2corTask, get_available_timestamps
from .meteoblue import MeteoblueVectorTask, MeteoblueRasterTask


__version__ = '0.10.2'



"""
Module for adding vector data from various sources

Credits:
Copyright (c) 2017-2021 Matej Aleksandrov, Matej Batič (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import logging

import boto3
import fiona
import geopandas as gpd
from eolearn.core import EOPatch, EOTask, FeatureTypeSet
from fiona.session import AWSSession

from sentinelhub import CRS, GeopediaFeatureIterator, SHConfig

LOGGER = logging.getLogger(__name__)


class _BaseVectorImportTask(EOTask):
    """Base Vector Import Task, implementing common methods
    """
    def __init__(self, feature, reproject=True, clip=False, config=None):
        """
        :param feature: A vector feature into which to import data
        :type feature: (FeatureType, str)
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :type reproject: bool, default = True
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :type clip: bool, default = False
        :param config: A configuration object with credentials
        :type config: SHConfig
        """
        self.feature = next(self._parse_features(feature, allowed_feature_types=FeatureTypeSet.VECTOR_TYPES)())
        self.config = config or SHConfig()
        self.reproject = reproject
        self.clip = clip

    def _reproject_and_clip(self, vectors, bbox):
        """Method to reproject and clip vectors to the EOPatch crs and bbox"""
        if not bbox and (self.reproject or self.clip):
            raise ValueError('To clip or reproject vector data, eopatch.bbox has to be defined!')

        if self.reproject:
            vectors = vectors.to_crs(bbox.crs.pyproj_crs())

        if self.clip:
            bbox_crs = bbox.crs.pyproj_crs()
            if vectors.crs != bbox_crs:
                raise ValueError('To clip, vectors should be in same CRS as EOPatch bbox!')

            extent = gpd.GeoSeries([bbox.geometry], crs=bbox_crs)
            vectors = gpd.clip(vectors, extent, keep_geom_type=True)

        return vectors

    def execute(self, eopatch=None, *, bbox=None):
        """
        :param eopatch: An existing EOPatch. If none is provided it will create a new one.
        :type eopatch: EOPatch
        :param bbox: A bounding box for which to load data. By default, if none is provided, it will take a bounding box
            of given EOPatch. If given EOPatch is not provided it will load the entire dataset.
        :return: An EOPatch with an additional vector feature
        :rtype: EOPatch
        """
        eopatch = eopatch or EOPatch()
        bbox = bbox or eopatch.bbox

        if not eopatch.bbox:
            eopatch.bbox = bbox

        vectors = self._load_vector_data(bbox)
        eopatch[self.feature] = self._reproject_and_clip(vectors, bbox)

        return eopatch


class VectorImportTask(_BaseVectorImportTask):
    """ A task for importing (Fiona readable) vector data files into an EOPatch
    """
    def __init__(self, feature, path, reproject=True, clip=False, config=None, **kwargs):
        """
        :param feature: A vector feature into which to import data
        :type feature: (FeatureType, str)
        :param path: A path to a dataset containing vector data. It can be either a local path or a path to s3 bucket
        :type path: str
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :type reproject: bool, default = True
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :type clip: bool, default = False
        :param config: A configuration object with AWS credentials (if not provided, ~/.aws/credentials will be used)
        :type config: SHConfig
        :param kwargs: Additional args that will be passed to `fiona.open` or `geopandas.read` calls (e.g. layer name)
        """
        self.path = path
        self.fiona_kwargs = kwargs
        self._aws_session = None
        self._dataset_crs = None

        super().__init__(feature=feature, reproject=reproject, clip=clip, config=config)

    @property
    def aws_session(self):
        """ Because the session object cannot be pickled this provides the session lazily (i.e. the first time it is
        needed)

        :return: A session for AWS services
        :rtype: AWSSession
        """
        if self._aws_session is None:
            session_credentials = {}
            if self.config and self.config.aws_access_key_id and self.config.aws_secret_access_key:
                session_credentials['aws_access_key_id'] = self.config.aws_access_key_id
                session_credentials['aws_secret_access_key'] = self.config.aws_secret_access_key

            boto_session = boto3.session.Session(**session_credentials)
            self._aws_session = AWSSession(boto_session)

        return self._aws_session

    @property
    def dataset_crs(self):
        """ Provides a CRS of dataset, it loads it lazily (i.e. the first time it is needed)

        :return: Dataset's CRS
        :rtype: CRS
        """
        if self._dataset_crs is None:
            if self.path.startswith('s3://'):
                with fiona.Env(session=self.aws_session):
                    self._read_crs()
            else:
                self._read_crs()

        return self._dataset_crs

    def _read_crs(self):
        """ Reads information about CRS from a dataset
        """
        with fiona.open(self.path, **self.fiona_kwargs) as features:
            self._dataset_crs = CRS(features.crs)

    def _load_vector_data(self, bbox):
        """ Loads vector data either from S3 or local path
        """
        bbox_bounds = bbox.transform_bounds(self.dataset_crs).geometry.bounds if bbox else None

        if self.path.startswith('s3://'):
            with fiona.Env(session=self.aws_session):
                with fiona.open(self.path, **self.fiona_kwargs) as features:
                    feature_iter = features if bbox_bounds is None else features.filter(bbox=bbox_bounds)

                    return gpd.GeoDataFrame.from_features(
                        feature_iter,
                        columns=list(features.schema['properties']) + ['geometry'],
                        crs=self.dataset_crs.pyproj_crs()
                    )

        return gpd.read_file(self.path, bbox=bbox_bounds, **self.fiona_kwargs)


class GeopediaVectorImportTask(_BaseVectorImportTask):
    """ A task for importing `Geopedia <https://geopedia.world>`__ features into EOPatch vector features
    """

    def __init__(self, feature, geopedia_table, reproject=True, clip=False, **kwargs):
        """
        :param feature: A vector feature into which to import data
        :type feature: (FeatureType, str)
        :param geopedia_table: A Geopedia table from which to retrieve features
        :type geopedia_table: str or int
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :type reproject: bool, default = True
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :type clip: bool, default = False
        :param kwargs: Additional args that will be passed to `GeopediaFeatureIterator`
        """
        self.geopedia_table = geopedia_table
        self.geopedia_kwargs = kwargs
        self.dataset_crs = None
        super().__init__(feature=feature, reproject=reproject, clip=clip)

    def _load_vector_data(self, bbox):
        """ Loads vector data from geopedia table
        """
        prepared_bbox = bbox.transform_bounds(CRS.POP_WEB) if bbox else None

        geopedia_iterator = GeopediaFeatureIterator(
            layer=self.geopedia_table,
            bbox=prepared_bbox,
            offset=0,
            gpd_session=None,
            config=self.config,
            **self.geopedia_kwargs
        )
        geopedia_features = list(geopedia_iterator)

        geometry = geopedia_features[0].get('geometry')
        if not geometry:
            raise ValueError(f'Geopedia table "{self.geopedia_table}" does not contain geometries!')

        self.dataset_crs = CRS(geometry['crs']['properties']['name'])  # always WGS84
        return gpd.GeoDataFrame.from_features(geopedia_features, crs=self.dataset_crs.pyproj_crs())


class GeoDBVectorImportTask(_BaseVectorImportTask):
    """ A task for importing vector data from `geoDB <https://eurodatacube.com/marketplace/services/edc_geodb>`__
    into EOPatch
    """
    def __init__(self, feature, geodb_client, geodb_collection, geodb_db,
                 reproject=True, clip=False, **kwargs):
        """
        :param feature: A vector feature into which to import data
        :type feature: (FeatureType, str)
        :param geodb_client: an instance of GeoDBClient
        :type geodb_client: xcube_geodb.core.geodb.GeoDBClient
        :param geodb_collection: The name of the collection to be queried
        :type geodb_collection: str
        :param geodb_db: The name of the database the collection resides in [current database]
        :type geodb_db: str
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :type reproject: bool, default = True
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :type clip: bool, default = False
        :param kwargs: Additional args that will be passed to `geodb_client.get_collection_by_bbox` call
            (e.g. where="id>-1", operator="and")
        """
        self.geodb_client = geodb_client
        self.geodb_db = geodb_db
        self.geodb_collection = geodb_collection
        self.geodb_kwargs = kwargs
        self._dataset_crs = None

        super().__init__(feature=feature, reproject=reproject, clip=clip)

    @property
    def dataset_crs(self):
        """ Provides a "crs" of dataset, loads it lazily (i.e. the first time it is needed)

        :return: Dataset's CRS
        :rtype: CRS
        """
        if self._dataset_crs is None:
            srid = self.geodb_client.get_collection_srid(
                collection=self.geodb_collection,
                database=self.geodb_db
            )
            self._dataset_crs = CRS(f'epsg:{srid}')

        return self._dataset_crs

    def _load_vector_data(self, bbox):
        """ Loads vector data from geoDB table
        """
        prepared_bbox = bbox.transform_bounds(self.dataset_crs).geometry.bounds if bbox else None

        if 'comparison_mode' not in self.geodb_kwargs:
            self.geodb_kwargs['comparison_mode'] = 'intersects'

        return self.geodb_client.get_collection_by_bbox(
            collection=self.geodb_collection,
            database=self.geodb_db,
            bbox=prepared_bbox,
            bbox_crs=self.dataset_crs.epsg,
            **self.geodb_kwargs
        )



"""
Module for adding data obtained from sentinelhub package to existing EOPatches

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import logging

import numpy as np
import rasterio.transform
import rasterio.warp
from sentinelhub import MimeType, CRS, GeopediaWmsRequest

from eolearn.core import EOTask, FeatureType
from eolearn.core.utilities import renamed_and_deprecated

LOGGER = logging.getLogger(__name__)


class AddGeopediaFeatureTask(EOTask):
    """
    Task for adding a feature from Geopedia to an existing EOPatch.

    At the moment the Geopedia supports only WMS requestes in EPSG:3857, therefore to add feature to EOPatch
    in arbitrary CRS and arbitrary service type the following steps are performed:
    * transform BBOX from EOPatch's CRS to EPSG:3857
    * get raster from Geopedia Request in EPSG:3857
    * vectorize the returned raster using rasterio
    * project vectorised raster back to EOPatch's CRS
    * rasterize back and add raster to EOPatch
    """

    def __init__(self, feature, layer, theme, raster_value, raster_dtype=np.uint8, no_data_val=0,
                 image_format=MimeType.PNG, mean_abs_difference=2):

        self.feature_type, self.feature_name = next(self._parse_features(feature)())

        self.raster_value = raster_value
        self.raster_dtype = raster_dtype
        self.no_data_val = no_data_val
        self.mean_abs_difference = mean_abs_difference

        self.layer = layer
        self.theme = theme

        self.image_format = image_format

    def _get_wms_request(self, bbox, size_x, size_y):
        """
        Returns WMS request.
        """
        bbox_3857 = bbox.transform(CRS.POP_WEB)

        return GeopediaWmsRequest(layer=self.layer,
                                  theme=self.theme,
                                  bbox=bbox_3857,
                                  width=size_x,
                                  height=size_y,
                                  image_format=self.image_format)

    def _get_wcs_request(self, bbox, size_x, size_y):
        """
        Returns WMS request.
        """
        raise NotImplementedError

    def _reproject(self, eopatch, src_raster):
        """
        Reprojects the raster data from Geopedia's CRS (POP_WEB) to EOPatch's CRS.
        """
        height, width = src_raster.shape

        dst_raster = np.ones((height, width), dtype=self.raster_dtype)

        src_bbox = eopatch.bbox.transform(CRS.POP_WEB)
        src_transform = rasterio.transform.from_bounds(*src_bbox, width=width, height=height)

        dst_bbox = eopatch.bbox
        dst_transform = rasterio.transform.from_bounds(*dst_bbox, width=width, height=height)

        rasterio.warp.reproject(src_raster, dst_raster,
                                src_transform=src_transform, src_crs=CRS.POP_WEB.ogc_string(),
                                src_nodata=0,
                                dst_transform=dst_transform, dst_crs=eopatch.bbox.crs.ogc_string(),
                                dst_nodata=self.no_data_val)

        return dst_raster

    def _to_binary_mask(self, array):
        """
        Returns binary mask (0 and raster_value)
        """
        # check where the transparency is not zero
        return (array[..., -1] > 0).astype(self.raster_dtype) * self.raster_value

    def _map_from_binaries(self, eopatch, dst_shape, request_data):
        """
        Each request represents a binary class which will be mapped to the scalar `raster_value`
        """
        if self.feature_name in eopatch[self.feature_type]:
            raster = eopatch[self.feature_type][self.feature_name].squeeze()
        else:
            raster = np.ones(dst_shape, dtype=self.raster_dtype) * self.no_data_val

        new_raster = self._reproject(eopatch, self._to_binary_mask(request_data))

        # update raster
        raster[new_raster != 0] = new_raster[new_raster != 0]

        return raster

    def _map_from_multiclass(self, eopatch, dst_shape, request_data):
        """
        `raster_value` is a dictionary specifying the intensity values for each class and the corresponding label value.

        A dictionary example for GLC30 LULC mapping is:
        raster_value = {'no_data': (0,[0,0,0,0]),
                        'cultivated land': (1,[193, 243, 249, 255]),
                        'forest': (2,[73, 119, 20, 255]),
                        'grassland': (3,[95, 208, 169, 255]),
                        'shrubland': (4,[112, 179, 62, 255]),
                        'water': (5,[154, 86, 1, 255]),
                        'wetland': (6,[244, 206, 126, 255]),
                        'tundra': (7,[50, 100, 100, 255]),
                        'artificial surface': (8,[20, 47, 147, 255]),
                        'bareland': (9,[202, 202, 202, 255]),
                        'snow and ice': (10,[251, 237, 211, 255])}
        """
        raster = np.ones(dst_shape, dtype=self.raster_dtype) * self.no_data_val

        for key in self.raster_value.keys():
            value, intensities = self.raster_value[key]
            raster[np.mean(np.abs(request_data - intensities), axis=-1) < self.mean_abs_difference] = value

        return self._reproject(eopatch, raster)

    def execute(self, eopatch):
        """
        Add requested feature to this existing EOPatch.
        """
        data_arr = eopatch[FeatureType.MASK]['IS_DATA']
        _, height, width, _ = data_arr.shape

        request = self._get_wms_request(eopatch.bbox, width, height)

        request_data, = np.asarray(request.get_data())

        if isinstance(self.raster_value, dict):
            raster = self._map_from_multiclass(eopatch, (height, width), request_data)
        elif isinstance(self.raster_value, (int, float)):
            raster = self._map_from_binaries(eopatch, (height, width), request_data)
        else:
            raise ValueError("Unsupported raster value type")

        if self.feature_type is FeatureType.MASK_TIMELESS and raster.ndim == 2:
            raster = raster[..., np.newaxis]

        eopatch[self.feature_type][self.feature_name] = raster

        return eopatch


@renamed_and_deprecated
class AddGeopediaFeature(AddGeopediaFeatureTask):
    """ A deprecated version of AddGeopediaFeatureTask
    """



""" Testing SentinelHubInputTask
"""
import datetime as dt
import os
import shutil
import dataclasses
from concurrent import futures
from typing import Optional, Any

import pytest
from pytest import approx
import numpy as np

from sentinelhub import BBox, CRS, DataCollection, SHConfig, Band, Unit
from eolearn.core import EOPatch, FeatureType, EOTask
from eolearn.io import (
    SentinelHubDemTask, SentinelHubEvalscriptTask, SentinelHubInputTask, SentinelHubSen2corTask,
    get_available_timestamps
)


@pytest.fixture(name='cache_folder')
def cache_folder_fixture():
    test_dir = os.path.dirname(os.path.realpath(__file__))
    cache_folder = os.path.join(test_dir, 'cache_test')

    if os.path.exists(cache_folder):
        shutil.rmtree(cache_folder)

    yield cache_folder

    shutil.rmtree(cache_folder)


@dataclasses.dataclass
class IoTestCase:
    name: str
    task: EOTask
    bbox: BBox
    time_interval: tuple
    feature: str = 'BANDS'
    feature_type: FeatureType = FeatureType.DATA
    data_size: Optional[int] = None
    timestamp_length: Optional[int] = None
    stats: Any = None


def calculate_stats(array):
    time, height, width, _ = array.shape
    edge1 = np.nanmean(array[int(time/2):, 0, 0, :])
    edge2 = np.nanmean(array[:max(int(time/2), 1), -1, -1, :])
    edge3 = np.nanmean(array[:, int(height/2), int(width/2), :])

    stats = np.round(np.array([edge1, edge2, edge3]), 4)

    return stats


@pytest.mark.sh_integration
class TestProcessingIO:
    """ Test cases for SentinelHubInputTask
    """
    size = (99, 101)
    bbox = BBox(bbox=[268892, 4624365, 268892+size[0]*10, 4624365+size[1]*10], crs=CRS.UTM_33N)
    time_interval = ('2017-12-15', '2017-12-30')
    maxcc = 0.8
    time_difference = dt.timedelta(minutes=60)
    max_threads = 3

    def test_s2l1c_float32_uint16(self, cache_folder):
        task = SentinelHubInputTask(
            bands_feature=(FeatureType.DATA, 'BANDS'),
            additional_data=[(FeatureType.MASK, 'dataMask')],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L1C,
            max_threads=self.max_threads,
            cache_folder=cache_folder
        )

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)
        bands = eopatch[(FeatureType.DATA, 'BANDS')]
        is_data = eopatch[(FeatureType.MASK, 'dataMask')]

        assert calculate_stats(bands) == approx([0.0233, 0.0468, 0.0252])

        width, height = self.size
        assert bands.shape == (4, height, width, 13)
        assert is_data.shape == (4, height, width, 1)
        assert len(eopatch.timestamp) == 4
        assert bands.dtype == np.float32

        assert os.path.exists(cache_folder)

        # change task's bans_dtype and run it again
        task.bands_dtype = np.uint16

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)
        bands = eopatch[(FeatureType.DATA, 'BANDS')]

        assert calculate_stats(bands) == approx([232.5769, 467.5385, 251.8654])

        assert bands.dtype == np.uint16

    def test_specific_bands(self):
        task = SentinelHubInputTask(
            bands_feature=(FeatureType.DATA, 'BANDS'),
            bands=["B01", "B02", "B03"],
            additional_data=[(FeatureType.MASK, 'dataMask')],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L1C,
            max_threads=self.max_threads
        )

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)
        bands = eopatch[(FeatureType.DATA, 'BANDS')]

        assert calculate_stats(bands) == approx([0.0648, 0.1193, 0.063])

        width, height = self.size
        assert bands.shape == (4, height, width, 3)

    def test_scl_only(self):
        """ Download just SCL, without any other bands
        """
        task = SentinelHubInputTask(
            bands_feature=None,
            additional_data=[(FeatureType.DATA, 'SCL')],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L2A,
            max_threads=self.max_threads
        )

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)
        scl = eopatch[(FeatureType.DATA, 'SCL')]

        width, height = self.size
        assert scl.shape == (4, height, width, 1)

    def test_single_scene(self):
        """ Download S2L1C bands and dataMask
        """
        task = SentinelHubInputTask(
            bands_feature=(FeatureType.DATA, 'BANDS'),
            additional_data=[(FeatureType.MASK, 'dataMask')],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L1C,
            max_threads=self.max_threads,
            single_scene=True,
            mosaicking_order="leastCC"
        )

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)
        bands = eopatch[(FeatureType.DATA, 'BANDS')]
        is_data = eopatch[(FeatureType.MASK, 'dataMask')]

        width, height = self.size
        assert bands.shape == (1, height, width, 13)
        assert is_data.shape == (1, height, width, 1)
        assert len(eopatch.timestamp) == 1

    def test_additional_data(self):
        task = SentinelHubInputTask(
            bands_feature=(FeatureType.DATA, 'BANDS'),
            bands=['B01', 'B02', 'B05'],
            additional_data=[
                (FeatureType.MASK, 'dataMask', 'IS_DATA'),
                (FeatureType.MASK, 'CLM'),
                (FeatureType.MASK, 'SCL'),
                (FeatureType.MASK, 'SNW'),
                (FeatureType.MASK, 'CLD'),
                (FeatureType.DATA, 'CLP'),
                (FeatureType.DATA, 'viewAzimuthMean', 'view_azimuth_mean'),
                (FeatureType.DATA, 'sunAzimuthAngles'),
                (FeatureType.DATA, 'sunZenithAngles')
            ],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L2A,
            max_threads=self.max_threads
        )

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)

        bands = eopatch[(FeatureType.DATA, 'BANDS')]
        is_data = eopatch[(FeatureType.MASK, 'IS_DATA')]
        clm = eopatch[(FeatureType.MASK, 'CLM')]
        scl = eopatch[(FeatureType.MASK, 'SCL')]
        snw = eopatch[(FeatureType.MASK, 'SNW')]
        cld = eopatch[(FeatureType.MASK, 'CLD')]
        clp = eopatch[(FeatureType.DATA, 'CLP')]
        view_azimuth_mean = eopatch[(FeatureType.DATA, 'view_azimuth_mean')]
        sun_azimuth_angles = eopatch[(FeatureType.DATA, 'sunAzimuthAngles')]
        sun_zenith_angles = eopatch[(FeatureType.DATA, 'sunZenithAngles')]

        assert calculate_stats(bands) == approx([0.027,  0.0243, 0.0162])

        width, height = self.size
        assert bands.shape == (4, height, width, 3)
        assert is_data.shape == (4, height, width, 1)
        assert is_data.dtype == bool
        assert clm.shape == (4, height, width, 1)
        assert clm.dtype == np.uint8
        assert scl.shape == (4, height, width, 1)
        assert snw.shape == (4, height, width, 1)
        assert cld.shape == (4, height, width, 1)
        assert clp.shape == (4, height, width, 1)
        assert view_azimuth_mean.shape == (4, height, width, 1)
        assert sun_azimuth_angles.shape == (4, height, width, 1)
        assert sun_zenith_angles.shape == (4, height, width, 1)
        assert len(eopatch.timestamp) == 4

    def test_aux_request_args(self):
        """ Download low resolution data with `PREVIEW` mode
        """
        task = SentinelHubInputTask(
            bands_feature=(FeatureType.DATA, 'BANDS'),
            resolution=260,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L1C,
            max_threads=self.max_threads,
            aux_request_args={'dataFilter': {'previewMode': 'PREVIEW'}}
        )

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)
        bands = eopatch[(FeatureType.DATA, 'BANDS')]

        assert bands.shape == (4, 4, 4, 13)
        assert calculate_stats(bands) == approx([0.0, 0.0493, 0.0277])

    def test_dem(self):
        task = SentinelHubDemTask(
            resolution=10,
            feature=(FeatureType.DATA_TIMELESS, 'DEM'),
            max_threads=3
        )

        eopatch = task.execute(bbox=self.bbox)
        dem = eopatch.data_timeless['DEM']

        width, height = self.size
        assert dem.shape == (height, width, 1)

    def test_dem_cop(self):
        task = SentinelHubDemTask(
            data_collection=DataCollection.DEM_COPERNICUS_30,
            resolution=10,
            feature=(FeatureType.DATA_TIMELESS, 'DEM_30'),
            max_threads=3
        )
        eopatch = task.execute(bbox=self.bbox)
        dem = eopatch.data_timeless['DEM_30']

        width, height = self.size
        assert dem.shape == (height, width, 1)

    def test_dem_wrong_feature(self):
        with pytest.raises(ValueError):
            SentinelHubDemTask(resolution=10, feature=(FeatureType.DATA, 'DEM'), max_threads=3)

    def test_sen2cor(self):
        task = SentinelHubSen2corTask(
            sen2cor_classification=['SCL', 'CLD'],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L2A,
            max_threads=self.max_threads
        )

        eopatch = task.execute(bbox=self.bbox, time_interval=self.time_interval)

        scl = eopatch[(FeatureType.MASK, 'SCL')]
        cld = eopatch[(FeatureType.DATA, 'CLD')]

        width, height = self.size
        assert scl.shape == (4, height, width, 1)
        assert cld.shape == (4, height, width, 1)

    def test_metadata(self):
        evalscript = """
            //VERSION=3

            function setup() {
                return {
                    input: [{
                        bands:["B02","dataMask"],
                        units: "DN"
                    }],
                    output:[
                    {
                        id:'bands',
                        bands: 1,
                        sampleType: SampleType.UINT16
                    }
                    ]
                }
            }

            function updateOutputMetadata(scenes, inputMetadata, outputMetadata) {
                outputMetadata.userData = { "metadata":  JSON.stringify(scenes) }
            }

            function evaluatePixel(sample) {
                return {
                    'bands': [sample.B02]
                };
            }
        """
        task = SentinelHubEvalscriptTask(
            evalscript=evalscript,
            data_collection=DataCollection.SENTINEL2_L1C,
            features=[(FeatureType.DATA, 'bands'),
                      (FeatureType.META_INFO, 'meta_info')],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            max_threads=self.max_threads
        )

        eop = task.execute(bbox=self.bbox, time_interval=self.time_interval)

        width, height = self.size
        assert eop.data['bands'].shape == (4, height, width, 1)
        assert len(eop.meta_info['meta_info']) == 4

    def test_multi_processing(self):
        task = SentinelHubInputTask(
            bands_feature=(FeatureType.DATA, 'BANDS'),
            bands=["B01", "B02", "B03"],
            additional_data=[(FeatureType.MASK, 'dataMask')],
            size=self.size,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            data_collection=DataCollection.SENTINEL2_L1C,
            max_threads=self.max_threads
        )

        time_intervals = [
            ('2017-01-01', '2017-01-30'),
            ('2017-02-01', '2017-02-28'),
            ('2017-03-01', '2017-03-30'),
            ('2017-04-01', '2017-04-30'),
            ('2017-05-01', '2017-05-30'),
            ('2017-06-01', '2017-06-30')
        ]

        with futures.ProcessPoolExecutor(max_workers=3) as executor:
            tasks = [executor.submit(task.execute, None, self.bbox, interval) for interval in time_intervals]
            eopatches = [task.result() for task in futures.as_completed(tasks)]

        array = np.concatenate([eop.data['BANDS'] for eop in eopatches], axis=0)

        width, height = self.size
        assert array.shape == (20, height, width, 3)

    def test_get_available_timestamps_with_missing_data_collection_service_url(self):
        collection = DataCollection.SENTINEL2_L1C.define_from('COLLECTION_WITHOUT_URL', service_url=None)
        timestamps = get_available_timestamps(
            bbox=self.bbox,
            config=SHConfig(),
            data_collection=collection,
            time_difference=self.time_difference,
            time_interval=self.time_interval,
            maxcc=self.maxcc
        )

        assert len(timestamps) == 4
        assert all(timestamp.tzinfo is not None for timestamp in timestamps)


@pytest.mark.sh_integration
class TestSentinelHubInputTaskDataCollections:
    """ Integration tests for all supported data collections
    """

    bbox = BBox(bbox=(-5.05, 48.0, -5.00, 48.05), crs=CRS.WGS84)
    bbox2 = BBox(bbox=(-72.2, -70.4, -71.6, -70.2), crs=CRS.WGS84)
    size = (50, 40)
    time_interval = ('2020-06-1', '2020-06-10')
    time_difference = dt.timedelta(minutes=60)
    data_feature = FeatureType.DATA, 'BANDS'
    mask_feature = FeatureType.MASK, 'dataMask'

    s3slstr_500m = DataCollection.SENTINEL3_SLSTR.define_from(
        'SENTINEL3_SLSTR_500m',
        bands=(
            Band('S2', (Unit.REFLECTANCE,), (np.float32,)),
            Band('S3', (Unit.REFLECTANCE,), (np.float32,)),
            Band('S6', (Unit.REFLECTANCE,), (np.float32,))
        )
    )
    s5p_co = DataCollection.SENTINEL5P.define_from(
        'SENTINEL5P_CO', bands=(Band('CO', (Unit.DN,), (np.float32,)),)
    )

    ndvi_evalscript = """
        //VERSION=3
        function setup() {
            return {
            input: [{
                bands: ["B04", "B08", "dataMask"],
                units: "DN"
            }],
            output: [
                { id:"ndvi", bands:1, sampleType: SampleType.FLOAT32 },
                { id:"dataMask", bands:1, sampleType: SampleType.UINT8 }
            ]
            }
        }
        function evaluatePixel(sample) {
        return {
            ndvi: [index(sample.B08, sample.B04)],
            dataMask: [sample.dataMask]};
        }
    """

    test_cases = [
        IoTestCase(
            name='Sentinel-2 L2A',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.SENTINEL2_L2A
            ),
            bbox=bbox,
            time_interval=time_interval,
            data_size=12,
            timestamp_length=2,
            stats=[0.4681, 0.6334, 0.7608]
        ),
        IoTestCase(
            name='Sentinel-2 L2A - NDVI evalscript',
            task=SentinelHubEvalscriptTask(
                features={
                    FeatureType.DATA: {'ndvi': 'NDVI-FEATURE'},
                    FeatureType.MASK: ['dataMask'],
                },
                evalscript=ndvi_evalscript,
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.SENTINEL2_L2A,
            ),
            feature='NDVI-FEATURE',
            bbox=bbox,
            time_interval=time_interval,
            data_size=1,
            timestamp_length=2,
            stats=[0.0036, 0.0158, 0.0088]
        ),
        IoTestCase(
            name='Landsat8',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.LANDSAT_OT_L1
            ),
            bbox=bbox,
            time_interval=time_interval,
            data_size=11,
            timestamp_length=1,
            stats=[48.7592, 48.726, 48.9168]
        ),
        IoTestCase(
            name='MODIS',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.MODIS
            ),
            bbox=bbox,
            time_interval=time_interval,
            data_size=7,
            timestamp_length=10,
            stats=[0.0073, 0.0101, 0.1448]
        ),
        IoTestCase(
            name='Sentinel-1 IW',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.SENTINEL1_IW
            ),
            bbox=bbox,
            time_interval=time_interval,
            data_size=2,
            timestamp_length=5,
            stats=[0.016, 0.0022, 0.0087]
        ),
        IoTestCase(
            name='Sentinel-1 IW ASCENDING',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.SENTINEL1_IW_ASC
            ),
            bbox=bbox,
            time_interval=time_interval,
            data_size=2,
            timestamp_length=1,
            stats=[0.0407, 0.0206, 0.0216]
        ),
        IoTestCase(
            name='Sentinel-1 EW DESCENDING',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.SENTINEL1_EW_DES
            ),
            bbox=bbox2,
            time_interval=time_interval,
            data_size=2,
            timestamp_length=1,
            stats=[np.nan, 0.1944, 0.3800]
        ),
        IoTestCase(
            name='Sentinel-3 OLCI',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=DataCollection.SENTINEL3_OLCI
            ),
            bbox=bbox,
            time_interval=time_interval,
            data_size=21,
            timestamp_length=11,
            stats=[0.317, 0.1946, 0.2884]
        ),
        # The following test doesn't work at the moment
        # IoTestCase(
        #     name='Sentinel-3 SLSTR 500m resolution',
        #     task=SentinelHubInputTask(
        #         bands_feature=data_feature,
        #         additional_data=[mask_feature],
        #         size=size,
        #         time_difference=time_difference,
        #         data_collection=s3slstr_500m
        #     ),
        #     bbox=bbox,
        #     time_interval=('2021-02-10', '2021-02-15'),
        #     data_size=3,
        #     timestamp_length=13,
        #     stats=[0.4236, 0.6339, 0.5117]
        # ),
        IoTestCase(
            name='Sentinel-5P',
            task=SentinelHubInputTask(
                bands_feature=data_feature,
                additional_data=[mask_feature],
                size=size,
                time_difference=time_difference,
                data_collection=s5p_co
            ),
            bbox=bbox,
            time_interval=('2020-06-1', '2020-06-1'),
            data_size=1,
            timestamp_length=1,
            stats=[0.0351, 0.034,  0.0351]
        )
    ]

    @pytest.mark.parametrize('test_case', test_cases)
    def test_data_collections(self, test_case):
        eopatch = test_case.task.execute(bbox=test_case.bbox, time_interval=test_case.time_interval)

        assert isinstance(eopatch, EOPatch), 'Expected return type is EOPatch'

        width, height = self.size
        data = eopatch[(test_case.feature_type, test_case.feature)]
        assert data.shape == (test_case.timestamp_length, height, width, test_case.data_size)

        timestamps = eopatch.timestamp
        assert all(timestamp.tzinfo is None for timestamp in timestamps), f'`tzinfo` present in timestamps {timestamps}'
        assert len(timestamps) == test_case.timestamp_length

        stats = calculate_stats(data)
        assert stats == approx(test_case.stats, nan_ok=True), f'Expected stats {test_case.stats}, got {stats}'



""" Input tasks that collect data from `Sentinel-Hub Process API
<https://docs.sentinel-hub.com/api/latest/api/process/>`__

Credits:
Copyright (c) 2019-2021 Matej Aleksandrov, Matej Batič, Matic Lubej, Jovan Višnjić (Sinergise)
Copyright (c) 2019-2021 Beno Šircelj

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import copy
import datetime as dt
import logging

import numpy as np
from eolearn.core import EOPatch, EOTask, FeatureType, FeatureTypeSet

from sentinelhub import (
    DataCollection, MimeType, SHConfig, SentinelHubCatalog, SentinelHubDownloadClient, SentinelHubRequest,
    bbox_to_dimensions, filter_times, parse_time_interval, Band, Unit
)

LOGGER = logging.getLogger(__name__)


def get_available_timestamps(bbox, config, data_collection, time_difference, time_interval=None, maxcc=None):
    """Helper function to search for all available timestamps, based on query parameters

    :param bbox: Bounding box
    :type bbox: BBox
    :param time_interval: Time interval to query available satellite data from
    :type time_interval: different input formats available (e.g. (str, str), or (datetime, datetime)
    :param data_collection: Source of requested satellite data.
    :type data_collection: DataCollection
    :param maxcc: Maximum cloud coverage, in ratio [0, 1], default is None
    :type maxcc: float
    :param time_difference: Minimum allowed time difference, used when filtering dates, None by default.
    :type time_difference: datetime.timedelta
    :param config: Sentinel Hub Config
    :type config: SHConfig
    :return: list of datetimes with available observations
    """
    query = None
    if maxcc and data_collection.has_cloud_coverage:
        if isinstance(maxcc, (int, float)) and (maxcc < 0 or maxcc > 1):
            raise ValueError('Maximum cloud coverage "maxcc" parameter should be a float on an interval [0, 1]')
        query = {'eo:cloud_cover': {'lte': int(maxcc * 100)}}

    fields = {'include': ['properties.datetime'], 'exclude': []}

    if data_collection.service_url:
        config = copy.copy(config)
        config.sh_base_url = data_collection.service_url
    catalog = SentinelHubCatalog(config=config)
    search_iterator = catalog.search(collection=data_collection, bbox=bbox, time=time_interval,
                                     query=query, fields=fields)

    all_timestamps = search_iterator.get_timestamps()
    filtered_timestamps = filter_times(all_timestamps, time_difference)

    if len(filtered_timestamps) == 0:
        raise ValueError(f'No available images for requested time range: {time_interval}')

    return filtered_timestamps


class SentinelHubInputBaseTask(EOTask):
    """ Base class for Processing API input tasks
    """

    def __init__(self, data_collection, size=None, resolution=None, cache_folder=None, config=None, max_threads=None):
        """
        :param data_collection: A collection of requested satellite data.
        :type data_collection: DataCollection
        :param size: Number of pixels in x and y dimension.
        :type size: tuple(int, int)
        :param resolution: Resolution in meters, passed as a single number or a tuple of two numbers -
            resolution in horizontal and resolution in vertical direction.
        :type resolution: float or (float, float)
        :param cache_folder: Path to cache_folder. If set to None (default) requests will not be cached.
        :type cache_folder: str
        :param config: An instance of SHConfig defining the service
        :type config: SHConfig or None
        :param max_threads: Maximum threads to be used when downloading data.
        :type max_threads: int
        """
        if (size is None) == (resolution is None):
            raise ValueError("Exactly one of the parameters 'size' and 'resolution' should be given.")

        self.size = size
        self.resolution = resolution
        self.config = config or SHConfig()
        self.max_threads = max_threads
        self.data_collection = DataCollection(data_collection)
        self.cache_folder = cache_folder

    def execute(self, eopatch=None, bbox=None, time_interval=None):
        """ Main execute method for the Process API tasks
        """

        eopatch = eopatch or EOPatch()

        self._check_and_set_eopatch_bbox(bbox, eopatch)
        size_x, size_y = self._get_size(eopatch)

        if time_interval:
            time_interval = parse_time_interval(time_interval)
            timestamp = self._get_timestamp(time_interval, eopatch.bbox)
        else:
            timestamp = eopatch.timestamp

        eop_timestamp = [time_point.replace(tzinfo=None) for time_point in timestamp]

        if eopatch.timestamp and eop_timestamp:
            self.check_timestamp_difference(timestamp, eopatch.timestamp)
        elif timestamp:
            eopatch.timestamp = eop_timestamp

        requests = self._build_requests(eopatch.bbox, size_x, size_y, timestamp, time_interval)
        requests = [request.download_list[0] for request in requests]

        LOGGER.debug('Downloading %d requests of type %s', len(requests), str(self.data_collection))
        client = SentinelHubDownloadClient(config=self.config)
        responses = client.download(requests, max_threads=self.max_threads)
        LOGGER.debug('Downloads complete')

        temporal_dim = len(timestamp) if timestamp else 1
        shape = temporal_dim, size_y, size_x
        self._extract_data(eopatch, responses, shape)

        eopatch.meta_info['size_x'] = size_x
        eopatch.meta_info['size_y'] = size_y
        if timestamp:  # do not overwrite time interval in case of timeless features
            eopatch.meta_info['time_interval'] = time_interval

        self._add_meta_info(eopatch)

        return eopatch

    def _get_size(self, eopatch):
        """Get the size (width, height) for the request either from inputs, or from the (existing) eopatch"""
        if self.size is not None:
            return self.size

        if self.resolution is not None:
            return bbox_to_dimensions(eopatch.bbox, self.resolution)

        if eopatch.meta_info and eopatch.meta_info.get('size_x') and eopatch.meta_info.get('size_y'):
            return eopatch.meta_info.get('size_x'), eopatch.meta_info.get('size_y')

        raise ValueError('Size or resolution for the requests should be provided!')

    def _add_meta_info(self, eopatch):
        """Add information to eopatch metadata"""
        if self.maxcc:
            eopatch.meta_info['maxcc'] = self.maxcc
        if self.time_difference:
            eopatch.meta_info['time_difference'] = self.time_difference

    @staticmethod
    def _check_and_set_eopatch_bbox(bbox, eopatch):
        if eopatch.bbox is None:
            if bbox is None:
                raise ValueError('Either the eopatch or the task must provide valid bbox.')
            eopatch.bbox = bbox
            return

        if bbox is None or eopatch.bbox == bbox:
            return
        raise ValueError('Either the eopatch or the task must provide bbox, or they must be the same.')

    @staticmethod
    def check_timestamp_difference(timestamp1, timestamp2):
        """ Raises an error if the two timestamps are not the same
        """
        error_msg = "Trying to write data to an existing eopatch with a different timestamp."
        if len(timestamp1) != len(timestamp2):
            raise ValueError(error_msg)

        for ts1, ts2 in zip(timestamp1, timestamp2):
            if ts1 != ts2:
                raise ValueError(error_msg)

    def _extract_data(self, eopatch, images, shape):
        """ Extract data from the received images and assign them to eopatch features
        """
        raise NotImplementedError("The _extract_data method should be implemented by the subclass.")

    def _build_requests(self, bbox, size_x, size_y, timestamp, time_interval):
        """ Build requests
        """
        raise NotImplementedError("The _build_requests method should be implemented by the subclass.")

    def _get_timestamp(self, time_interval, bbox):
        """ Get the timestamp array needed as a parameter for downloading the images
        """
        raise NotImplementedError("The _get_timestamp method should be implemented by the subclass.")


class SentinelHubEvalscriptTask(SentinelHubInputBaseTask):
    """ Process API task to download data using evalscript
    """

    def __init__(self, features=None, evalscript=None, data_collection=None, size=None, resolution=None,
                 maxcc=None, time_difference=None, cache_folder=None,
                 max_threads=None, config=None, mosaicking_order=None, aux_request_args=None):
        """
        :param features: Features to construct from the evalscript.
        :param evalscript: Evalscript for the request. Beware that all outputs from SentinelHub services should be named
            and should have the same name as corresponding feature
        :type evalscript: str
        :param data_collection: Source of requested satellite data.
        :type data_collection: DataCollection
        :param size: Number of pixels in x and y dimension.
        :type size: tuple(int, int)
        :param resolution: Resolution in meters, passed as a single number or a tuple of two numbers -
            resolution in horizontal and resolution in vertical direction.
        :type resolution: float or (float, float)
        :param maxcc: Maximum cloud coverage, a float in interval [0, 1]
        :type maxcc: float
        :param time_difference: Minimum allowed time difference, used when filtering dates, None by default.
        :type time_difference: datetime.timedelta
        :param cache_folder: Path to cache_folder. If set to None (default) requests will not be cached.
        :type cache_folder: str
        :param config: An instance of SHConfig defining the service
        :type config: SHConfig or None
        :param max_threads: Maximum threads to be used when downloading data.
        :type max_threads: int
        :param mosaicking_order: Mosaicking order, which has to be either 'mostRecent', 'leastRecent' or 'leastCC'.
        :type mosaicking_order: str
        :param aux_request_args: a dictionary with auxiliary information for the input_data part of the SH request
        :type aux_request_args: dict
        """
        super().__init__(data_collection=data_collection, size=size, resolution=resolution, cache_folder=cache_folder,
                         config=config, max_threads=max_threads)

        self.features = self._parse_and_validate_features(features)
        self.responses = self._create_response_objects()

        if not evalscript:
            raise ValueError('evalscript parameter must not be missing/empty')
        self.evalscript = evalscript

        if maxcc and isinstance(maxcc, (int, float)) and (maxcc < 0 or maxcc > 1):
            raise ValueError('maxcc should be a float on an interval [0, 1]')

        self.maxcc = maxcc
        self.time_difference = time_difference or dt.timedelta(seconds=1)
        self.mosaicking_order = mosaicking_order
        self.aux_request_args = aux_request_args

    def _parse_and_validate_features(self, features):
        if not features:
            raise ValueError('features must be defined')

        allowed_features = FeatureTypeSet.RASTER_TYPES.union({FeatureType.META_INFO})
        _features = list(self._parse_features(features, allowed_feature_types=allowed_features, new_names=True)())

        ftr_data_types = set(ft for ft, _, _ in _features if not ft.is_meta())
        if all(ft.is_timeless() for ft in ftr_data_types) or all(ft.is_time_dependent() for ft in ftr_data_types):
            return _features

        raise ValueError('Cannot mix time dependent and timeless requests!')

    def _create_response_objects(self):
        """ Construct SentinelHubRequest output_responses from features
        """
        responses = []
        for feat_type, feat_name, _ in self.features:
            if feat_type.is_raster():
                responses.append(SentinelHubRequest.output_response(feat_name, MimeType.TIFF))
            elif feat_type.is_meta():
                responses.append(SentinelHubRequest.output_response('userdata', MimeType.JSON))
            else:
                # should not happen as features have already been validated
                raise ValueError(f'{feat_type} not supported!')

        return responses

    def _get_timestamp(self, time_interval, bbox):
        """ Get the timestamp array needed as a parameter for downloading the images
        """
        if any(feat_type.is_timeless() for feat_type, _, _ in self.features if feat_type.is_raster()):
            return []

        return get_available_timestamps(bbox=bbox, time_interval=time_interval, data_collection=self.data_collection,
                                        maxcc=self.maxcc, time_difference=self.time_difference, config=self.config)

    def _build_requests(self, bbox, size_x, size_y, timestamp, time_interval):
        """ Build requests
        """
        if timestamp:
            dates = [(date - self.time_difference, date + self.time_difference) for date in timestamp]
        else:
            dates = [parse_time_interval(time_interval, allow_undefined=True)] if time_interval else [None]

        return [self._create_sh_request(date, bbox, size_x, size_y) for date in dates]

    def _create_sh_request(self, time_interval, bbox, size_x, size_y):
        """ Create an instance of SentinelHubRequest
        """
        return SentinelHubRequest(
            evalscript=self.evalscript,
            input_data=[SentinelHubRequest.input_data(
                data_collection=self.data_collection,
                mosaicking_order=self.mosaicking_order,
                time_interval=time_interval,
                maxcc=self.maxcc,
                other_args=self.aux_request_args
            )],
            responses=self.responses,
            bbox=bbox,
            size=(size_x, size_y),
            data_folder=self.cache_folder,
            config=self.config
        )

    def _extract_data(self, eopatch, data_responses, shape):
        """ Extract data from the received images and assign them to eopatch features
        """
        # pylint: disable=arguments-renamed
        if len(self.features) == 1:
            ftype, fname, _ = self.features[0]
            extension = 'json' if ftype.is_meta() else 'tif'
            data_responses = [{f'{fname}.{extension}': data} for data in data_responses]

        for ftype, fname, new_fname in self.features:
            if ftype.is_meta():
                data = [data['userdata.json'] for data in data_responses]

            elif ftype.is_time_dependent():
                data = np.asarray([data[f"{fname}.tif"] for data in data_responses])
                data = data[..., np.newaxis] if data.ndim == 3 else data

            else:
                data = np.asarray(data_responses[0][f"{fname}.tif"])[..., np.newaxis]

            eopatch[ftype][new_fname] = data

        return eopatch


class SentinelHubInputTask(SentinelHubInputBaseTask):
    """ Process API input task that loads 16bit integer data and converts it to a 32bit float feature.
    """
    # pylint: disable=too-many-arguments
    DTYPE_TO_SAMPLE_TYPE = {
        bool: 'SampleType.UINT8',
        np.uint8: 'SampleType.UINT8',
        np.uint16: 'SampleType.UINT16',
        np.float32: 'SampleType.FLOAT32',
    }

    def __init__(self, data_collection=None, size=None, resolution=None, bands_feature=None, bands=None,
                 additional_data=None, evalscript=None, maxcc=None, time_difference=None, cache_folder=None,
                 max_threads=None, config=None, bands_dtype=None, single_scene=False, mosaicking_order=None,
                 aux_request_args=None):
        """
        :param data_collection: Source of requested satellite data.
        :type data_collection: DataCollection
        :param size: Number of pixels in x and y dimension.
        :type size: tuple(int, int)
        :param resolution: Resolution in meters, passed as a single number or a tuple of two numbers -
            resolution in horizontal and resolution in vertical direction.
        :type resolution: float or (float, float)
        :param bands_feature: Target feature into which to save the downloaded images.
        :type bands_feature: tuple(sentinelhub.FeatureType, str)
        :param bands: An array of band names. If not specified it will download all bands specified for a given data
            collection.
        :type bands: list[str]
        :param additional_data: A list of additional data to be downloaded, such as SCL, SNW, dataMask, etc.
        :type additional_data: list[tuple(sentinelhub.FeatureType, str)]
        :param evalscript: An optional parameter to override an evascript that is generated by default
        :type evalscript: str or None
        :param maxcc: Maximum cloud coverage.
        :type maxcc: float
        :param time_difference: Minimum allowed time difference, used when filtering dates, None by default.
        :type time_difference: datetime.timedelta
        :param cache_folder: Path to cache_folder. If set to None (default) requests will not be cached.
        :type cache_folder: str
        :param config: An instance of SHConfig defining the service
        :type config: SHConfig or None
        :param max_threads: Maximum threads to be used when downloading data.
        :type max_threads: int
        :param bands_dtype: output type of the bands array, if set to None the default is used
        :type bands_dtype: type or None
        :param single_scene: If true, the service will compute a single image for the given time interval using
            mosaicking.
        :type single_scene: bool
        :param mosaicking_order: Mosaicking order, which has to be either 'mostRecent', 'leastRecent' or 'leastCC'.
        :type mosaicking_order: str
        :param aux_request_args: a dictionary with auxiliary information for the input_data part of the SH request
        :type aux_request_args: dict
        """
        super().__init__(data_collection=data_collection, size=size, resolution=resolution, cache_folder=cache_folder,
                         config=config, max_threads=max_threads)
        self.evalscript = evalscript
        self.maxcc = maxcc
        self.time_difference = time_difference or dt.timedelta(seconds=1)
        self.single_scene = single_scene
        self.bands_dtype = bands_dtype
        self.mosaicking_order = mosaicking_order
        self.aux_request_args = aux_request_args

        self.bands_feature = None
        self.requested_bands = []
        if bands_feature:
            self.bands_feature = next(self._parse_features(bands_feature, allowed_feature_types=[FeatureType.DATA])())
            if bands:
                self.requested_bands = self._parse_requested_bands(bands, self.data_collection.bands)
            else:
                self.requested_bands = list(self.data_collection.bands)

        self.requested_additional_bands = []
        if additional_data is not None:
            additional_data = list(self._parse_features(additional_data, new_names=True)())
            additional_bands = [band for _, band, _ in additional_data]
            parsed_bands = self._parse_requested_bands(additional_bands, self.data_collection.metabands)
            self.requested_additional_bands = parsed_bands

        self.additional_data = additional_data

    def _parse_requested_bands(self, bands, available_bands):
        """ Checks that all requested bands are available and returns the band information for further processing """
        requested_bands = []
        band_info_dict = {band_info.name: band_info for band_info in available_bands}
        for band_name in bands:
            if band_name in band_info_dict:
                requested_bands.append(band_info_dict[band_name])
            elif self.data_collection.is_batch or self.data_collection.is_byoc:
                requested_bands.append(Band(band_name, (Unit.DN,), (np.float32,)))
            else:
                raise ValueError(
                    f'Data collection {self.data_collection} does not have specifications for {band_name}.'
                    f'Available bands are {[band.name for band in self.data_collection.bands]} and meta-bands'
                    f'{[band.name for band in self.data_collection.metabands]}'
                )
        return requested_bands

    def generate_evalscript(self):
        """ Generate the evalscript to be passed with the request, based on chosen bands
        """
        evalscript = """
            //VERSION=3

            function setup() {{
                return {{
                    input: [{{
                        bands: [{bands}],
                        units: [{units}]
                    }}],
                    output: [
                        {outputs}
                    ]
                }}
            }}

            function evaluatePixel(sample) {{
                return {{ {samples} }}
            }}
        """

        bands, units, outputs, samples = [], [], [], []
        for band in self.requested_bands + self.requested_additional_bands:
            unit_choice = 0  # use default units
            if band in self.requested_bands and self.bands_dtype is not None:
                if self.bands_dtype not in band.output_types:
                    raise ValueError(
                        f'Band {band.name} only supports output types {band.output_types} but `bands_dtype` is set to '
                        f'{self.bands_dtype}. To use default types set `bands_dtype` to None.'
                    )
                unit_choice = band.output_types.index(self.bands_dtype)

            sample_type = SentinelHubInputTask.DTYPE_TO_SAMPLE_TYPE[band.output_types[unit_choice]]

            bands.append(f'"{band.name}"')
            units.append(f'"{band.units[unit_choice].value}"')
            samples.append(f'{band.name}: [sample.{band.name}]')
            outputs.append(f'{{ id: "{band.name}", bands: 1, sampleType: {sample_type} }}')

        evalscript = evalscript.format(
            bands=', '.join(bands), units=', '.join(units), outputs=', '.join(outputs), samples=', '.join(samples)
        )

        return evalscript

    def _get_timestamp(self, time_interval, bbox):
        """ Get the timestamp array needed as a parameter for downloading the images
        """
        if self.single_scene:
            return [time_interval[0]]

        return get_available_timestamps(bbox=bbox, time_interval=time_interval, data_collection=self.data_collection,
                                        maxcc=self.maxcc, time_difference=self.time_difference, config=self.config)

    def _build_requests(self, bbox, size_x, size_y, timestamp, time_interval):
        """ Build requests
        """
        if self.single_scene:
            dates = [parse_time_interval(time_interval)]
        else:
            dates = [(date - self.time_difference, date + self.time_difference) for date in timestamp]

        return [self._create_sh_request(date1, date2, bbox, size_x, size_y) for date1, date2 in dates]

    def _create_sh_request(self, date_from, date_to, bbox, size_x, size_y):
        """ Create an instance of SentinelHubRequest
        """
        responses = [SentinelHubRequest.output_response(band.name, MimeType.TIFF)
                     for band in self.requested_bands + self.requested_additional_bands]

        return SentinelHubRequest(
            evalscript=self.evalscript or self.generate_evalscript(),
            input_data=[
                SentinelHubRequest.input_data(
                    data_collection=self.data_collection,
                    time_interval=(date_from, date_to),
                    mosaicking_order=self.mosaicking_order,
                    maxcc=self.maxcc,
                    other_args=self.aux_request_args
                )
            ],
            responses=responses,
            bbox=bbox,
            size=(size_x, size_y),
            data_folder=self.cache_folder,
            config=self.config
        )

    def _extract_data(self, eopatch, images, shape):
        """ Extract data from the received images and assign them to eopatch features
        """
        if len(self.requested_bands) + len(self.requested_additional_bands) == 1:
            # if only one band is requested the response is not a tar so we reshape it
            only_band = (self.requested_bands + self.requested_additional_bands)[0]
            images = [{only_band.name + '.tif': image} for image in images]

        if self.additional_data:
            self._extract_additional_features(eopatch, images, shape)

        if self.bands_feature:
            self._extract_bands_feature(eopatch, images, shape)

        return eopatch

    def _extract_additional_features(self, eopatch, images, shape):
        """ Extracts additional features from response into an EOPatch
        """
        for (ftype, _, new_name), band_info in zip(self.additional_data, self.requested_additional_bands):
            tifs = [tar[band_info.name + '.tif'] for tar in images]
            eopatch[ftype, new_name] = self._extract_array(tifs, 0, shape, band_info.output_types[0])

    def _extract_bands_feature(self, eopatch, images, shape):
        """ Extract the bands feature arrays and concatenate them along the last axis
        """
        processed_bands = []
        for band_info in self.requested_bands:
            tifs = [tar[band_info.name + '.tif'] for tar in images]
            dtype = self.bands_dtype or band_info.output_types[0]
            processed_bands.append(self._extract_array(tifs, 0, shape, dtype))

        eopatch[self.bands_feature] = np.concatenate(processed_bands, axis=-1)

    @staticmethod
    def _extract_array(tifs, idx, shape, dtype):
        """ Extract a numpy array from the received tifs
        """
        feature_arrays = (np.atleast_3d(img)[..., idx] for img in tifs)
        return np.asarray(list(feature_arrays), dtype=dtype).reshape(*shape, 1)


class SentinelHubDemTask(SentinelHubEvalscriptTask):
    """
    Adds DEM data (one of the `collections <https://docs.sentinel-hub.com/api/latest/data/dem/#deminstance>`__) to
        DATA_TIMELESS EOPatch feature.
    """

    def __init__(self, feature=None, data_collection=DataCollection.DEM, **kwargs):
        if feature is None:
            feature = (FeatureType.DATA_TIMELESS, 'dem')
        elif isinstance(feature, str):
            feature = (FeatureType.DATA_TIMELESS, feature)

        feature_type, feature_name = feature
        if feature_type.is_time_dependent():
            raise ValueError("DEM feature should be timeless!")

        band = data_collection.bands[0]

        evalscript = f"""
            //VERSION=3

            function setup() {{
                return {{
                    input: [{{
                        bands: ["{band.name}"],
                        units: ["{band.units[0].value}"]
                    }}],
                    output: {{
                        id: "{feature_name}",
                        bands: 1,
                        sampleType: SampleType.UINT16
                    }}
                }}
            }}

            function evaluatePixel(sample) {{
                return {{ {feature_name}: [sample.{band.name}] }}
            }}
        """

        super().__init__(evalscript=evalscript, features=[feature], data_collection=data_collection, **kwargs)


class SentinelHubSen2corTask(SentinelHubInputTask):
    """
    Adds SCL (scene classification), CLD (cloud probability) or SNW (snow probability) (or any their combination)
    Sen2Cor classification results to EOPatch's MASK or DATA feature. The feature is added to MASK (SCL) or
    DATA (CLD, SNW) feature types of EOPatch. The feature names are set to be SCL, CLD or SNW.

    Sen2Cor's scene classification (SCL) contains 11 classes with the following values and meanings:
       * 1 - SC_SATURATED_DEFECTIVE
       * 2 - SC_DARK_FEATURE_SHADOW
       * 3 - SC_CLOUD_SHADOW
       * 4 - VEGETATION
       * 5 - NOT-VEGETATED
       * 6 - WATER
       * 7 - SC_CLOUD_LOW_PROBA / UNCLASSIFIED
       * 8 - SC_CLOUD_MEDIUM_PROBA
       * 9 - CLOUD_HIGH_PROBABILITY
       * 10 - THIN_CIRRUS
       * 11 - SNOW
    """

    def __init__(self, sen2cor_classification, data_collection=DataCollection.SENTINEL2_L2A, **kwargs):
        """
        :param sen2cor_classification: "SCL" (scene classification), "CLD" (cloud probability) or "SNW"
            (snow probability) masks to be retrieved. Also a list of their combination (e.g. ["SCL","CLD"])
        :type sen2cor_classification: str or [str]
        :param kwargs: Additional arguments that will be passed to the `SentinelHubInputTask`
        """
        # definition of possible types and target features
        classification_types = {
            'SCL': FeatureType.MASK,
            'CLD': FeatureType.DATA,
            'SNW': FeatureType.DATA
        }

        if isinstance(sen2cor_classification, str):
            sen2cor_classification = [sen2cor_classification]

        for s2c in sen2cor_classification:
            if s2c not in classification_types:
                raise ValueError(f'Unsupported Sen2Cor classification type: {s2c}. '
                                 f'Possible types are: {classification_types}!')

        if data_collection != DataCollection.SENTINEL2_L2A:
            raise ValueError('Sen2Cor classification layers are only available on Sentinel-2 L2A data.')

        features = [(classification_types[s2c], s2c) for s2c in sen2cor_classification]
        super().__init__(additional_data=features, data_collection=data_collection, **kwargs)



"""
Module with tasks that provide data from Meteoblue services

To use tasks from this module you have to use Python >= `3.7` and install METEOBLUE package extension:

.. code-block::

    pip install eo-learn-io[METEOBLUE]

Credits:
Copyright (c) 2021 Patrick Zippenfenig (Meteoblue)
Copyright (c) 2021 Matej Aleksandrov (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import datetime as dt
from typing import Optional

import dateutil.parser
import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.geometry
from sentinelhub import parse_time_interval, serialize_time, Geometry, CRS
try:
    import meteoblue_dataset_sdk
    from meteoblue_dataset_sdk.caching import FileCache
except ImportError:
    pass

from eolearn.core import EOTask, EOPatch


class BaseMeteoblueTask(EOTask):
    """ A base task implementing the logic that is common for all Meteoblue tasks
    """
    def __init__(self, feature, apikey: str, query: dict, units: dict = None,
                 time_difference: dt.timedelta = dt.timedelta(minutes=30),
                 cache_folder: Optional[str] = None, cache_max_age: int = 604800):
        """
        :param feature: A feature in which Meteoblue data will be stored
        :type feature: (FeatureType, str)
        :param apikey: Meteoblue API key
        :type apikey: str
        :param query: Meteoblue dataset API query definition
        :type query: dict
        :param units: Meteoblue dataset API units definition
        :type units: dict
        :param time_difference: The size of a time interval around each timestamp for which data will be collected. It
            is used only in a combination with ``time_interval`` parameter from ``execute`` method.
        :type time_difference: datetime.timedelta
        :param cache_folder: Path to cache_folder. If set to None (default) requests will not be cached.
        :type cache_folder: str
        :param cache_max_age: Maximum age in seconds to use a cached result. Default 1 week.
        :type cache_max_age: int
        """
        self.feature = next(self._parse_features(feature)())
        cache = None
        if cache_folder:
            cache = FileCache(path=cache_folder, max_age=cache_max_age)

        self.client = meteoblue_dataset_sdk.Client(apikey=apikey, cache=cache)
        self.query = query
        self.units = units
        self.time_difference = time_difference

    @staticmethod
    def _prepare_bbox(eopatch, bbox):
        """ Prepares a bbox from input parameters
        """
        if not eopatch.bbox and not bbox:
            raise ValueError('Bounding box is not provided')
        if eopatch.bbox and bbox and eopatch.bbox != bbox:
            raise ValueError('Provided eopatch.bbox and bbox are not the same')

        return bbox or eopatch.bbox

    def _prepare_time_intervals(self, eopatch, time_interval):
        """ Prepare a list of time intervals for which data will be collected from Meteoblue services
        """
        if not eopatch.timestamp and not time_interval:
            raise ValueError('Time interval should either be defined with eopatch.timestamp of time_interval parameter')

        if time_interval:
            start_time, end_time = serialize_time(parse_time_interval(time_interval))
            return [f'{start_time}/{end_time}']

        timestamps = eopatch.timestamp
        time_intervals = []
        for timestamp in timestamps:
            start_time = timestamp - self.time_difference
            end_time = timestamp + self.time_difference

            start_time, end_time = serialize_time((start_time, end_time))
            time_interval = f'{start_time}/{end_time}'

            time_intervals.append(time_interval)

        return time_intervals

    def _get_data(self, query):
        """ It should return an output feature object and a list of timestamps
        """
        raise NotImplementedError

    def execute(self, eopatch=None, *, bbox=None, time_interval=None):
        """ Execute method that adds new Meteoblue data into an EOPatch

        :param eopatch: An EOPatch in which data will be added. If not provided a new EOPatch will be created.
        :type eopatch: EOPatch or None
        :param bbox: A bounding box of a request. Should be provided if eopatch parameter is not provided.
        :type bbox: BBox or None
        :param time_interval: An interval for which data should be downloaded. If not provided then timestamps from
            provided eopatch will be used.
        :type time_interval: (dt.datetime, dt.datetime) or (str, str) or None
        """
        eopatch = eopatch or EOPatch()
        eopatch.bbox = self._prepare_bbox(eopatch, bbox)
        time_intervals = self._prepare_time_intervals(eopatch, time_interval)

        bbox = eopatch.bbox
        geometry = Geometry(bbox.geometry, bbox.crs).transform(CRS.WGS84)
        geojson = shapely.geometry.mapping(geometry.geometry)
        query = {
            'units': self.units,
            'geometry': geojson,
            'format': 'protobuf',
            'timeIntervals': time_intervals,
            'queries': [self.query],
        }
        result_data, result_timestamp = self._get_data(query)

        if not eopatch.timestamp and result_timestamp:
            eopatch.timestamp = result_timestamp

        eopatch[self.feature] = result_data
        return eopatch


class MeteoblueVectorTask(BaseMeteoblueTask):
    """ Obtains weather data from Meteoblue services as a vector feature

    The data is obtained as a VECTOR feature in a ``geopandas.GeoDataFrame`` where columns include latitude, longitude,
    timestamp and a columns for each weather variable. All data is downloaded from the
    `Meteoblue dataset API <https://docs.meteoblue.com/en/apis/environmental-data/dataset-api>`__.

    A Meteoblue API key is required to retrieve data.
    """
    def _get_data(self, query):
        """ Provides a GeoDataFrame with information about weather control points and an empty list of timestamps
        """
        result = self.client.querySync(query)
        dataframe = meteoblue_to_dataframe(result)
        geometry = gpd.points_from_xy(dataframe.Longitude, dataframe.Latitude)
        crs = CRS.WGS84.pyproj_crs()
        gdf = gpd.GeoDataFrame(dataframe, geometry=geometry, crs=crs)
        return gdf, []


class MeteoblueRasterTask(BaseMeteoblueTask):
    """ Obtains weather data from Meteoblue services as a raster feature

    It returns a 4D numpy array with dimensions (time, height, width, weather variables) which should be stored as a
    DATA feature. Data is resampled to WGS84 plate carrée to a specified resolution using the
    `Meteoblue dataset API <https://docs.meteoblue.com/en/apis/environmental-data/dataset-api>`__.

    A Meteoblue API key is required to retrieve data.
    """
    def _get_data(self, query):
        """ Return a 4-dimensional numpy array of shape (time, height, width, weather variables) and a list of
        timestamps
        """
        result = self.client.querySync(query)

        data = meteoblue_to_numpy(result)
        timestamps = _meteoblue_timestamps_from_geometry(result.geometries[0])
        return data, timestamps


def meteoblue_to_dataframe(result) -> pd.DataFrame:
    """ Transform a Meteoblue dataset API result to a dataframe

    :param result: A response of Meteoblue API
    :type result: Dataset_pb2.DatasetApiProtobuf
    :returns: A dataframe with columns TIMESTAMP, Longitude, Latitude and aggregation columns
    """
    geometry = result.geometries[0]
    code_names = [f'{code.code}_{code.level}_{code.aggregation}' for code in geometry.codes]

    if not geometry.timeIntervals:
        return pd.DataFrame(columns=['TIMESTAMP', 'Longitude', 'Latitude'] + code_names)

    dataframes = []
    for index, time_interval in enumerate(geometry.timeIntervals):
        timestamps = _meteoblue_timestamps_from_time_interval(time_interval)

        n_locations = len(geometry.lats)
        n_timesteps = len(timestamps)

        dataframe = pd.DataFrame({
            'TIMESTAMP': np.tile(timestamps, n_locations),
            'Longitude': np.repeat(geometry.lons, n_timesteps),
            'Latitude': np.repeat(geometry.lats, n_timesteps)
        })

        for code, code_name in zip(geometry.codes, code_names):
            dataframe[code_name] = np.array(code.timeIntervals[index].data)

        dataframes.append(dataframe)

    return pd.concat(dataframes, ignore_index=True)


def meteoblue_to_numpy(result) -> np.ndarray:
    """ Transform a Meteoblue dataset API result to a dataframe

    :param result: A response of Meteoblue API
    :type result: Dataset_pb2.DatasetApiProtobuf
    :returns: A 4D numpy array with shape (time, height, width, weather variables)
    """
    geometry = result.geometries[0]

    n_locations = len(geometry.lats)
    n_codes = len(geometry.codes)
    n_time_intervals = len(geometry.timeIntervals)
    geo_ny = geometry.ny
    geo_nx = geometry.nx

    # Meteoblue data is using dimensions (n_variables, n_time_intervals, ny, nx, n_timesteps)
    # Individual time intervals may have different number of timesteps (not a dimension)
    # Therefore we have to first transpose each code individually and then transpose everything again
    def map_code(code):
        """ Transpose a single code
        """
        code_data = np.array(list(map(lambda t: t.data, code.timeIntervals)))

        code_n_timesteps = code_data.size // n_locations // n_time_intervals
        code_data = code_data.reshape((n_time_intervals, geo_ny, geo_nx, code_n_timesteps))

        return code_data.transpose((0, 3, 1, 2))

    data = np.array(list(map(map_code, geometry.codes)))

    n_timesteps = data.size // n_locations // n_codes
    data = data.reshape((n_codes, n_timesteps, geo_ny, geo_nx))

    return data.transpose((1, 2, 3, 0))


def _meteoblue_timestamps_from_geometry(geometry_pb):
    """ Transforms a protobuf geometry object into a list of datetime objects
    """
    return list(pd.core.common.flatten(map(_meteoblue_timestamps_from_time_interval, geometry_pb.timeIntervals)))


def _meteoblue_timestamps_from_time_interval(timestamp_pb):
    """ Transforms a protobuf timestamp object into a list of datetime objects
    """
    if timestamp_pb.timestrings:
        # Time intervals like weekly data, return an `array of strings` as timestamps
        # For time indications like `20200801T0000-20200802T235959` we only return the first date as datetime
        return list(map(_parse_timestring, timestamp_pb.timestrings))

    # Regular time intervals return `start, end and stride` as a time axis
    # We convert it into an array of daytime
    time_range = range(timestamp_pb.start, timestamp_pb.end, timestamp_pb.stride)
    return list(map(dt.date.fromtimestamp, time_range))


def _parse_timestring(timestring):
    """ A helper method to parse specific timestrings obtained from Meteoblue service
    """
    if '-' in timestring:
        timestring = timestring.split('-')[0]
    return dateutil.parser.parse(timestring)
